/*
 * Copyright (C) 2006-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file user_list.c User handling for para_server. */

#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "fd.h"
#include "string.h"
#include "list.h"
#include "user_list.h"

static struct list_head user_list;

/*
 * Fill the list of users known to para_server.
 *
 * Populates a linked list of all users in \a user_list_file.  Returns on
 * success, calls exit() on errors.
 */
static void populate_user_list(char *user_list_file)
{
	int ret = -E_USERLIST;
	FILE *file_ptr = fopen(user_list_file, "r");

	if (!file_ptr)
		goto err;
	for (;;) {
		int num;
		char line[255];
		/* keyword, name, key, perms */
		char w[255], n[255], k[255], p[255], tmp[4][255];
		struct user *u;
		RSA *rsa;

		ret = para_fgets(line, MAXLINE, file_ptr);
		if (ret <= 0)
			break;
		if (sscanf(line,"%200s %200s %200s %200s", w, n, k, p) < 3)
			continue;
		if (strcmp(w, "user"))
			continue;
		PARA_DEBUG_LOG("found entry for user %s\n", n);
		ret = get_rsa_key(k, &rsa, LOAD_PUBLIC_KEY);
		if (ret < 0) {
			PARA_NOTICE_LOG("skipping entry for user %s: %s\n", n,
				para_strerror(-ret));
			continue;
		}
		u = para_malloc(sizeof(*u));
		u->name = para_strdup(n);
		u->rsa = rsa;
		u->perms = 0;
		num = sscanf(p, "%200[A-Z_],%200[A-Z_],%200[A-Z_],%200[A-Z_]",
			tmp[0], tmp[1], tmp[2], tmp[3]);
		PARA_DEBUG_LOG("found %i perm entries\n", num);
		while (num > 0) {
			num--;
			if (!strcmp(tmp[num], "VSS_READ"))
				u->perms |= VSS_READ;
			else if (!strcmp(tmp[num], "VSS_WRITE"))
				u->perms |= VSS_WRITE;
			else if (!strcmp(tmp[num], "AFS_READ"))
				u->perms |= AFS_READ;
			else if (!strcmp(tmp[num], "AFS_WRITE"))
				u->perms |= AFS_WRITE;
			else /* unknown permission */
				PARA_WARNING_LOG("ignoring unknown permission: %s\n",
					tmp[num]);
		}
		para_list_add(&u->node, &user_list);
	}
	fclose(file_ptr);
	if (ret >= 0)
		return;
err:
	PARA_EMERG_LOG("%s\n", para_strerror(-ret));
	exit(EXIT_FAILURE);
}

/**
 * Initialize the list of users allowed to connect to to para_server.
 *
 * \param user_list_file The file containing access information.
 *
 * If this function is called for the second time, the contents of the
 * previous call are discarded, i.e. the user list is reloaded.
 */
void init_user_list(char *user_list_file)
{
	struct user *u, *tmp;
	static int initialized;

	if (initialized) {
		list_for_each_entry_safe(u, tmp, &user_list, node) {
			list_del(&u->node);
			free(u->name);
			rsa_free(u->rsa);
			free(u);
		}
	} else
		INIT_LIST_HEAD(&user_list);
	initialized = 1;
	populate_user_list(user_list_file);
}

/**
 * Lookup a user in the user list.
 *
 * \param name The name of the user.
 *
 * \return A pointer to the corresponding user struct if the user was found, \p
 * NULL otherwise.
 */
struct user *lookup_user(const char *name)
{
	struct user *u;
	list_for_each_entry(u, &user_list, node) {
		if (strcmp(u->name, name))
			continue;
		return u;
	}
	return NULL;
}
